#
#  This file is part of the Connection-Set Algebra (CSA).
#  Copyright (C) 2010,2011,2012 Mikael Djurfeldt
#
#  CSA is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  CSA is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import numpy as _numpy
import matplotlib
import matplotlib.pyplot as _plt

from . import elementary

# This function was autogenerated by boilerplate.py.  Do not edit as
# changes will be lost
def inverseGray():
    '''
    set the default colormap to gray and apply to current image if any.
    See help(colormaps) for more information
    '''
    _plt.rc('image', cmap='gray_r')
    im = _plt.gci()

    if im is not None:
        im.set_cmap(_plt.cm.gray_r)
    _plt.draw_if_interactive()

def show (cset, N0 = 30, N1 = None):
    N1 = N0 if N1 == None else N1
    _plt.clf ()
    _plt.axis ('equal')
    a = _numpy.zeros ((N0, N1))
    for (i, j) in elementary.cross (range (N0), range (N1)) * cset:
        a[i,j] += 1.0
    _plt.imshow (a, interpolation='nearest')
    _plt.show ()

def gplotsel2d (g, cset, source = elementary.N, target = elementary.N, N0 = 900, N1 = None, value = None, range=[], lines = True):
    N1 = N0 if N1 == None else N1
    _plt.clf ()
    _plt.axis ('equal')
    gplot2d (g, N1, color = 'grey', show = False)
    cset = elementary.cross (source, target) * cset
    N = len (cset)
    if lines:
        marker = 'ro-'
    else:
        marker = 'ro'
    if elementary.arity (cset):
        if value != None:
            if range:
                normalize = matplotlib.colors.Normalize (*range)
            else:
                normalize = matplotlib.colors.Normalize ()
                normalize.autoscale ([v[value] for (i, j, v) in cset.c])
            cmap = matplotlib.cm.get_cmap ()
            for (i, j, v) in cset.c:
                color = cmap (normalize (v[value]))
                _plt.plot ([g (i)[0], g (j)[0]], [g (i)[1], g (j)[1]], \
                           marker, color = color, mfc = color)
        else:
            for (i, j, v) in cset.c:
                _plt.plot ([g (i)[0], g (j)[0]], [g (i)[1], g (j)[1]], marker)
    else:
        for (i, j) in cset:
            _plt.plot ([g (i)[0], g (j)[0]], [g (i)[1], g (j)[1]], marker)
    _plt.show ()

def gplot2d (g, N, color = None, show = True):
    if show:
        _plt.clf ()
        _plt.axis ('equal')
    x = []
    y = []
    for i in range (0, N):
        pos = g (i)
        x.append (pos[0])
        y.append (pos[1])
    if color != None:
        _plt.plot (x, y, 'o', color = color)
    else:
        _plt.plot (x, y, 'bo')
    if show:
        _plt.show ()

#del numpy, plt



#
#  This file is part of the Connection-Set Algebra (CSA).
#  Copyright (C) 2010,2011,2012 Mikael Djurfeldt
#
#  CSA is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  CSA is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import copy

from . import intervalset
from . import valueset

from .csaobject import *

# This is the fundamental connection-set class
# which is also the base class for masks
#
class CSet (CSAObject):
    tag = 'cset'
    
    def __init__ (self, mask, *valueSets):
        CSAObject.__init__ (self, "icset");
        self._mask = mask
        self.valueSets = list (valueSets)
        self.arity = len (self.valueSets)

    def repr (self, name = None):
        if self.arity:
            if not name:
                name = self.name
            vreprs = []
            for k in range (self.arity):
                v = self.value (k)
                if isinstance (v, CSAObject):
                    vreprs += ", %s" % v.repr ()
                else:
                    vreprs += ", %s" % v
            return "%s (%s%s)" % (self.name, self.mask (), "".join (vreprs))
        else:
            if self.mask () == self:
                return self.name
            else:
                return "%s (%s)" % (self.name, self.mask ())

    def mask (self):
        #*fixme* remove this condition?
        if self._mask == None:
            self._mask = self.makeMask ()
        return self._mask

    def value (self, k):
        if self.valueSets[k] == None:
            self.valueSets[k] = self.makeValueSet (k)
        return self.valueSets[k]

    def makeValueSet (self, k):
        if isFinite (self.mask ()):
            return self.makeFiniteValueSet (k, self.mask ().bounds ())
        raise RuntimeError ("don't know how to return value set for this connection-set")

    def makeFiniteValueSet (self, k, bounds):
        raise RuntimeError ("don't know how to return value set for this connection-set")

    def __len__ (self):
        return len (self.mask ())

    def __iter__ (self):
        # this code is used for full connection sets
        if isFinite (self.mask ()):
            state = State ()
            obj = self.startIteration (state)
            (low0, high0, low1, high1) = self.bounds ()
            return obj.iterator (low0, high0, low1, high1, state)
        else:
            raise RuntimeError ('attempt to retrieve iterator over infinite connection-set')

    def bounds (self):
        return self.mask ().bounds ()

    def startIteration (self, state):
        obj = copy.copy (self)
        obj._mask = self.mask ().startIteration (state)
        return obj

    def iterator (self, low0, high0, low1, high1, state):
        for (i, j) in self._mask.iterator (low0, high0, low1, high1, state):
            yield (i, j, [ v (i, j) for v in self.valueSets ])

    def multisetSum (self, other):
        return CSetMultisetSum (self, other)

    def intersection (self, other):
        assert isinstance (other, Mask), 'expected Mask operand'
        return SubCSet (self,
                        self.mask ().intersection (other), *self.valueSets)

    def difference (self, other):
        assert isinstance (other, Mask), 'expected Mask operand'
        return SubCSet (self, self.mask ().difference (other), *self.valueSets)


# This is the connection-set wrapper class which has as its only purpose
# to wrap non mask connection-sets so that the same code can implement
# connection-sets of different arity.  Some type dispatch is also done here.
#
class ConnectionSet (CSAObject):
    def __init__ (self, c):
        CSAObject.__init__ (self, "cset")
        self.c = c

    def repr (self):
        return self.c.repr ()
        
    def __len__ (self):
        return len (self.c)

    def __iter__ (self):
        return ConnectionSet.iterators[self.c.arity] (self)

    def iter0 (self):
        assert False, 'Should not have executed ConnectionSet.iter0'

    def iter1 (self):
        for (i, j, vs) in iter (self.c):
            (v0,) = vs
            yield (i, j, v0)

    def iter2 (self):
        for (i, j, vs) in iter (self.c):
            (v0, v1) = vs
            yield (i, j, v0, v1)

    def iter3 (self):
        for (i, j, vs) in iter (self.c):
            (v0, v1, v2) = vs
            yield (i, j, v0, v1, v2)

    def __add__ (self, other):
        if isNumber (other):
            return ConnectionSet (self.c.addScalar (other))
        else:
            return ConnectionSet (self.c.multisetSum (coerceCSet (other)))

    def __radd__ (self, other):
        return self.__add__ (other)

    def __sub__ (self, other):
        if isNumber (other):
            return ConnectionSet (self.c.addScalar (- other))
        else:
            return ConnectionSet (self.c.difference (coerceCSet (other)))

    def __rsub__ (self, other):
        return ConnectionSet (self.c.__neg__ ().addScalar (other))
    
    def __mul__ (self, other):
        if isNumber (other):
            return ConnectionSet (self.c.mulScalar (other))
        else:
            return ConnectionSet (self.c.intersection (coerceCSet (other)))

    def __rmul__ (self, other):
        return self.__mul__ (other)

ConnectionSet.iterators = [ ConnectionSet.iter0, \
                            ConnectionSet.iter1, \
                            ConnectionSet.iter2, \
                            ConnectionSet.iter3 ]


# Some helper functions

def source (x):
    return x[0]

def target (x):
    return x[1]

def isNumber (x):
    return isinstance (x, (int, float, complex))

def coerceCSet (obj):
    if isinstance (obj, list):
        return ExplicitMask (obj)
    elif isinstance (obj, ConnectionSet):
        return obj.c
    assert isinstance (obj, Mask), 'expected connection-set'
    return obj

def valueSet (obj):
    return valueset.QuotedValueSet (obj)

def coerceValueSet (obj):
    if callable (obj):
        return obj
    else:
        return valueSet (obj)

def isFinite (x):
    return isinstance (x, Finite)

def isEmpty (x):
    iterator = iter (x.mask ())
    try:
        next (iterator)
        return False
    except StopIteration:
        return True

def transpose (obj):
    return obj.transpose ()


# This is the fundamental mask class
#
class Mask (CSet):
    def __init__ (self):
        CSet.__init__ (self, self)

    def __len__ (self):
        N = 0
        for c in self:
            N += 1
        return N

    def __iter__ (self):
        raise RuntimeError ('attempt to retrieve iterator over infinite mask')

    def __add__ (self, other):
        return self.multisetSum (other)

    def __sub__ (self, other):
        return self.difference (other)

    def __mul__ (self, other):
        if isinstance (other, Mask):
            return self.intersection (other)
        elif isinstance (other, list):
            return self.intersection (ExplicitMask (other))
        elif isinstance (other, ConnectionSet):
            return other.__mul__ (self)
        else:
            return NotImplemented

    def __rmul__ (self, other):
        if isinstance (other, list):
            return self.intersection (ExplicitMask (other))
        else:
            return NotImplemented

    def __invert__ (self):
        return self.complement ()

    def transpose (self):
        assert isFinite (self), \
               'transpose currently only supports finite masks'
        return TransposedMask (self)

    def shift (self, M, N):
        return shiftedMask (self, M, N)
    
    def startIteration (self, state):
        # default action:
        return self

    def iterator (self, low0, high0, low1, high1, state):
        return NotImplemented

    def multisetSum (self, other):
        if isFinite (self) and isFinite (other):
            return FiniteMaskMultisetSum (self, other)
        else:
            return MaskMultisetSum (self, other)

    def intersection (self, other):
        # IntervalSetMask implements a specialized version of intersection
        if isinstance (other, IntervalSetMask):
            return other.intersection (self)
        # Generate Finite instances if either operand is finite
        elif isFinite (self):
            return FiniteMaskIntersection (self, other)
        elif isFinite (other):
            return FiniteMaskIntersection (other, self)
        else:
            return MaskIntersection (self, other)

    def complement (self):
        return MaskComplement (self)

    def difference (self, other):
        return MaskDifference (self, other)


class Finite (object):
    def bounds (self):
        return NotImplemented

    def maxBounds (self, b1, b2):
        return (min (b1[0], b2[0]), max (b1[1], b2[1]),
                min (b1[2], b2[2]), max (b1[3], b2[3]))

    def __iter__ (self):
        state = State ()
        obj = self.startIteration (state)
        (low0, high0, low1, high1) = self.bounds ()
        return obj.iterator (low0, high0, low1, high1, state)


class FiniteMask (Finite, Mask):
    def __init__ (self):
        Mask.__init__ (self)
        self.low0 = 0
        self.high0 = 0
        self.low1 = 0
        self.high1 = 0

    def bounds (self):
        return (self.low0, self.high0, self.low1, self.high1)

    def isBoundedBy (self, low0, high0, low1, high1):
        return low0 > self.low0 or high0 < self.high0 \
               or low1 > self.low1 or high1 < self.high1

# not used
class NoParIterator ():
    def __init__ (self):
        self.subIterator = False

    def iterator (self, low0, high0, low1, high1, state):
        print(low0, high0, low1, high1)
        if not self.subIterator:
            self.subIterator = self.noParIterator (state)
            self.lastC = next (self.subIterator)
        c = self.lastC
        while c[1] < low1:
            c = next (self.subIterator)
        while c[1] < high1:
            j = c[1]
            while c[1] == j and c[0] < low0:
                c = next (self.subIterator)
            while c[1] == j and c[0] < high0:
                yield c
                c = next (self.subIterator)
            while c[1] == j:
                c = next (self.subIterator)
        self.lastC = c


class BinaryMask (BinaryCSAObject, Mask):
    def __init__ (self, operator, op1, op2, precedence):
        Mask.__init__ (self)
        BinaryCSAObject.__init__ (self, operator, op1, op2, precedence)

    def startIteration (self, state):
        obj = copy.copy (self)
        obj.op1 = self.op1.startIteration (state)
        obj.op2 = self.op2.startIteration (state)
        return obj


class MaskIntersection (BinaryMask):
    def __init__ (self, op1, op2):
        BinaryMask.__init__ (self, '*', op1, op2, 1)

    def iterator (self, low0, high0, low1, high1, state):
        iter1 = self.op1.iterator (low0, high0, low1, high1, state)
        iter2 = self.op2.iterator (low0, high0, low1, high1, state)
        (i1, j1) = next (iter1)
        (i2, j2) = next (iter2)
        while True:
            if (j1, i1) < (j2, i2):
                (i1, j1) = next (iter1)
            elif (j2, i2) < (j1, i1):
                (i2, j2) = next (iter2)
            else:
                yield (i1, j1)
                (i1, j1) = next (iter1)
                (i2, j2) = next (iter2)


class FiniteMaskIntersection (Finite, MaskIntersection):
    def __init__ (self, op1, op2):
        assert isFinite (op1)
        MaskIntersection.__init__ (self, op1, op2)

    def bounds (self):
        return self.op1.bounds ()


class MaskMultisetSum (BinaryMask):
    def __init__ (self, op1, op2):
        BinaryMask.__init__ (self, "+", op1, op2, 0)

    def iterator (self, low0, high0, low1, high1, state):
        iter1 = self.op1.iterator (low0, high0, low1, high1, state)
        iter2 = self.op2.iterator (low0, high0, low1, high1, state)
        try:
            (i1, j1) = next (iter1)
        except StopIteration:
            (i2, j2) = next (iter2)
            while True:
                yield (i2, j2)
                (i2, j2) = next (iter2)
        try:
            (i2, j2) = next (iter2)
        except StopIteration:
            while True:
                yield (i1, j1)
                (i1, j1) = next (iter1)
        while True:
            i1s = i1
            j1s = j1
            while (j1, i1) <= (j2, i2):
                yield (i1, j1)
                try:
                    (i1, j1) = next (iter1)
                except StopIteration:
                    while True:
                        yield (i2, j2)
                        (i2, j2) = next (iter2)
            while (j2, i2) <= (j1s, i1s):
                yield (i2, j2)
                try:
                    (i2, j2) = next (iter2)
                except StopIteration:
                    while True:
                        yield (i1, j1)
                        (i1, j1) = next (iter1)


class FiniteMaskMultisetSum (Finite, MaskMultisetSum):
    def __init__ (self, op1, op2):
        assert isFinite (op1) and isFinite (op2)
        MaskMultisetSum.__init__ (self, op1, op2)

    def bounds (self):
        return self.maxBounds (self.op1.bounds (), self.op2.bounds ())


class MaskDifference (BinaryMask):
    def __init__ (self, op1, op2):
        BinaryMask.__init__ (self, "-", op1, op2, 0)

    def iterator (self, low0, high0, low1, high1, state):
        iter1 = self.op1.iterator (low0, high0, low1, high1, state)
        iter2 = self.op2.iterator (low0, high0, low1, high1, state)
        (i1, j1) = next (iter1)
        (i2, j2) = next (iter2)
        while True:
            if (j1, i1) < (j2, i2):
                yield (i1, j1)
                (i1, j1) = next (iter1)
                continue
            elif (i1, j1) == (i2, j2):
                (i1, j1) = next (iter1)
            try:
                (i2, j2) = next (iter2)
            except StopIteration:
                while True:
                    yield (i1, j1)
                    (i1, j1) = next (iter1)


def cmpPostOrder (c0, op1):
    return  ((c0[1], c0[0]) > (op1[1], op1[0])) -  ((c0[1], c0[0]) < (op1[1], op1[0]))


def cmp_to_key(mycmp):
    'Convert a cmp= function into a key= function'
    class K:
        def __init__(self, obj, *args):
            self.obj = obj
        def __lt__(self, other):
            return mycmp(self.obj, other.obj) < 0
        def __gt__(self, other):
            return mycmp(self.obj, other.obj) > 0
        def __eq__(self, other):
            return mycmp(self.obj, other.obj) == 0
        def __le__(self, other):
            return mycmp(self.obj, other.obj) <= 0
        def __ge__(self, other):
            return mycmp(self.obj, other.obj) >= 0
        def __ne__(self, other):
            return mycmp(self.obj, other.obj) != 0
    return K


class ExplicitMask (FiniteMask):
    def __init__ (self, connections):
        FiniteMask.__init__ (self)
        self.connections = list (connections)
        self.connections.sort (key=cmp_to_key(cmpPostOrder))
        if connections:
            self.low0 = min ((i for (i, j) in self.connections))
            self.high0 = max ((i for (i, j) in self.connections)) + 1
            self.low1 = self.connections[0][1]
            self.high1 = self.connections[-1][1] + 1

    def __len__ (self):
        return len (self.connections)

    def iterator (self, low0, high0, low1, high1, state):
        if not self.isBoundedBy (low0, high0, low1, high1):
            return iter (self.connections)
        else:
            return self.boundedIterator (low0, high0, low1, high1, state)

    def boundedIterator (self, low0, high0, low1, high1, state):
        iterator = iter (self.connections)
        (i, j) = next (iterator)
        while j < low1:
            (i, j) = next (iterator)
        while j < high1:
            if low0 <= i and i < high0:
                yield (i, j)
            (i, j) = next (iterator)


class IntervalSetMask (Mask):
    tag = 'cross'
    
    def __init__ (self, set0, set1):
        Mask.__init__ (self)
        self.set0 = set0
        self.set1 = set1

    @staticmethod
    def _sets_to_repr (set0, set1):
        return 'cross(%s, %s)' % (set0.repr (), set1.repr ())
    
    def repr (self):
        return self._sets_to_repr (self.set0, self.set1)

    def __contains__ (self, c):
        return c[0] in self.set0 and c[1] in self.set1

    def transpose (self):
        return IntervalSetMask (self.set1, self.set0)

    def shift (self, M, N):
        return IntervalSetMask (self.set0.shift (M), self.set1.shift (N))

    def iterator (self, low0, high0, low1, high1, state):
        iterator1 = self.set1.intervalIterator ()
        i1 = next (iterator1)
        while i1[1] < low1:
            i1 = next (iterator1)
        while i1[0] < high1:
            for j in range (max (i1[0], low1), min (i1[1] + 1, high1)):
                iterator0 = self.set0.intervalIterator ()
                try:
                    i0 = next (iterator0)
                    while i0[1] < low0:
                        i0 = next (iterator0)
                    if i0[1] < high0:
                        for i in range (max (i0[0], low0), i0[1] + 1):
                            yield (i, j)
                        i0 = next (iterator0)
                        while i0[1] < high0:
                            for i in range (i0[0], i0[1] + 1):
                                yield (i, j)
                            i0 = next (iterator0)
                        for i in range (i0[0], min (i0[1] + 1, high0)):
                            yield (i, j)
                    else:
                        for i in range (max (i0[0], low0), min (i0[1] + 1, high0)):
                            yield (i, j)
                except StopIteration:
                    pass
            i1 = next (iterator1)

    def intersection (self, other):
        if isinstance (other, IntervalSetMask):
            set0 = self.set0.intersection (other.set0)
            set1 = self.set1.intersection (other.set1)
            return intervalSetMask (set0, set1)
        else:
            return ISetBoundedMask (self.set0, self.set1, other)

    def multisetSum (self, other):
        if isinstance (other, IntervalSetMask):
            if not self.set0.intersection (other.set0) \
               or not self.set1.intersection (other.set1):
                set0 = self.set0.union (other.set0)
                set1 = self.set1.union (other.set1)
                return intervalSetMask (set0, set1)
            else:
                raise RuntimeError ('sums of overlapping IntervalSetMask:s not yet supported')
        else:
            return FiniteMask.multisetSum (self, other)

    @staticmethod
    def _sets_to_xml (set0, set1):
        return CSAObject.apply (IntervalSetMask.tag, set0, set1)

    def _to_xml (self):
        return self._sets_to_xml (self.set0, self.set1)


class FiniteISetMask (FiniteMask, IntervalSetMask):
    def __init__ (self, set0, set1):
        FiniteMask.__init__ (self)
        IntervalSetMask.__init__ (self, set0, set1)
        if self.set0 and self.set1:
            self.low0 = self.set0.min ()
            self.high0 = self.set0.max () + 1
            self.low1 = self.set1.min ()
            self.high1 = self.set1.max () + 1

    def __len__ (self):
        return len (self.set0) * len (self.set1)

    def transpose (self):
        return FiniteISetMask (self.set1, self.set0)

    def shift (self, M, N):
        return FiniteISetMask (self.set0.shift (M), self.set1.shift (N))

    def iterator (self, low0, high0, low1, high1, state):
        if not self.isBoundedBy (low0, high0, low1, high1):
            return self.simpleIterator ()
        else:
            return IntervalSetMask.iterator (self, low0, high0, low1, high1, state)

    def simpleIterator (self):
        for j in self.set1:
            for i in self.set0:
                yield (i, j)


class FiniteSourcesISetMask (IntervalSetMask):
    def __init__ (self, set0, set1):
        IntervalSetMask.__init__ (self, set0, set1)

    def transpose (self):
        return FiniteTargetsISetMask (self.set1, self.set0)

    def shift (self, M, N):
        return FiniteSourcesISetMask (self.set0.shift (M), \
                                      self.set1.shift (N))


class FiniteTargetsISetMask (IntervalSetMask):
    def __init__ (self, set0, set1):
        IntervalSetMask.__init__ (self, set0, set1)

    def transpose (self):
        return FiniteSourcesISetMask (self.set1, self.set0)

    def shift (self, M, N):
        return FiniteTargetsISetMask (self.set0.shift (M), \
                                      self.set1.shift (N))


def intervalSetMask (set0, set1):
    set0 = set0 if isinstance (set0, intervalset.IntervalSet) \
           else intervalset.IntervalSet (set0)
    set1 = set1 if isinstance (set1, intervalset.IntervalSet) \
           else intervalset.IntervalSet (set1)
    if set0.finite ():
        if set1.finite ():
            return FiniteISetMask (set0, set1)
        else:
            return FiniteSourcesISetMask (set0, set1)
    else:
        if set1.finite ():
            return FiniteTargetsISetMask (set0, set1)
        else:
            return IntervalSetMask (set0, set1)

CSAObject.tag_map[CSA + IntervalSetMask.tag] = (intervalSetMask, 2)


class ISetBoundedMask (FiniteMask):
    def __init__ (self, set0, set1, mask):
        FiniteMask.__init__ (self)
        self.precedence = 1
        self.set0 = set0
        self.set1 = set1
        self.subMask = mask
        inf = intervalset.infinity
        if isFinite (mask):
            (low0, high0, low1, high1) = mask.bounds ()
        else:
            (low0, high0, low1, high1) = (0, inf, 0, inf)
        if self.set0 and self.set1:
            self.low0 = max (self.set0.min (), low0)
            if self.set0.finite ():
                self.high0 = min (self.set0.max () + 1, high0)
            else:
                self.high0 = high0
            self.low1 = max (self.set1.min (), low1)
            if self.set1.finite ():
                self.high1 = min (self.set1.max () + 1, high1)
            else:
                self.high1 = high1
        assert self.high0 != inf and self.high1 != inf, 'infinite ISetBoundedMask:s currently not supported'

    def startIteration (self, state):
        obj = copy.copy (self)
        obj.subMask = self.subMask.startIteration (state)
        return obj

    def iterator (self, low0, high0, low1, high1, state):
        if not self.isBoundedBy (low0, high0, low1, high1):
            return self.simpleIterator (state)
        else:
            return self.boundedIterator (low0, high0, low1, high1, state)

    def simpleIterator (self, state):
        for i1 in self.set1.intervalIterator ():
            for i0 in self.set0.intervalIterator ():
                for e in self.subMask.iterator (i0[0], i0[1] + 1,
                                                i1[0], i1[1] + 1,
                                                state):
                    yield e

    def boundedIterator (self, low0, high0, low1, high1, state):
        iterator1 = self.set1.intervalIterator ()
        i1 = next (iterator1)
        while i1[1] < low1:
            i1 = next (iterator1)
        while i1[0] < high1:
            i1 = (max (i1[0], low1), min (i1[1], high1 - 1))
            iterator0 = self.set0.intervalIterator ()
            try:
                i0 = next (iterator0)
                while i0[1] < low0:
                    i0 = next (iterator0)
                if i0[1] < high0:
                    for e in self.subMask.iterator (max (i0[0], low0),
                                                    i0[1] + 1,
                                                    i1[0], i1[1] + 1,
                                                    state):
                        yield e
                    i0 = next (iterator0)
                    while i0[1] < high0:
                        for e in self.subMask.iterator (i0[0], i0[1] + 1,
                                                        i1[0], i1[1] + 1,
                                                        state):
                            yield e
                        i0 = next (iterator0)
                        for e in self.subMask.iterator (i0[0],
                                                        min (i0[1] + 1, high0),
                                                        i1[0], i1[1] + 1,
                                                        state):
                            yield e
                else:
                        for e in self.subMask.iterator (max (i0[0], low0),
                                                        min (i0[1] + 1, high0),
                                                        i1[0], i1[1] + 1,
                                                        state):
                            yield e
            except StopIteration:
                pass
            i1 = next (iterator1)

    def repr (self):
        return '%s*%s' % (IntervalSetMask._sets_to_repr (self.set0, self.set1),
                          self.subMask._repr_as_op2 (self.precedence))

    def _to_xml (self):
        return E ('apply',
                  E ('times'),
                  IntervalSetMask._sets_to_xml (self.set0, self.set1),
                  self.subMask._to_xml ())

# The ExplicitCSet captures the original value sets before coercion.
# It is used in the implementation of the "cset" constructor.
#
class ExplicitCSet (CSet):
    def __init__ (self, mask, *valueSets):
        if isinstance (mask, list):
            mask = ExplicitMask (mask)
        self.originalValueSets = valueSets
        CSet.__init__ (self, mask, *list (map (coerceValueSet, valueSets)))

    def value (self, k):
        return self.originalValueSets[k]


# SubCSet is used in the cases where a new CSet can be created by
# an operation on the mask.
#
class SubCSet (CSet):
    def __init__ (self, cset, mask, *valueSets):
        CSet.__init__ (self, mask, *valueSets)
        self.subCSet = cset

    def value (self, k):
        if self.valueSets[k] == None:
            self.valueSets[k] = self.makeValueSet (k)
        # defer to subCSet in case it is an ExplicitCSet
        return self.subCSet.value (k)

    def makeValueSet (self, k):
        if isFinite (self.mask ()):
            bounds = self.mask ().bounds ()
            return self.subCSet.makeFiniteValueSet (k, bounds)
        else:
            return self.subCSet.makeValueSet (k)


class BinaryCSet (BinaryCSAObject, CSet):
    def __init__ (self, operator, op1, op2):
        CSet.__init__ (self, None, *[ None for v in op1.valueSets ])
        self.name = operator
        self.op1 = op1
        self.op2 = op2
        self.valueSetMap = None

    def makeFiniteValueSet (self, k, bounds):
        if self.valueSetMap == None:
            self.valueSetMap = self.makeValueSetMap (bounds)
        return lambda i, j: self.valueSetMap[(i, j)][k]

    def makeValueSetMap (self, bounds):
        m = {}
        state = State ()
        obj = self.startIteration (state)
        (low0, high0, low1, high1) = bounds
        for (i, j, v) in obj.iterator (low0, high0, low1, high1, state):
            m[(i, j)] = v
        return m


class BinaryCSets (BinaryCSet):
    def __init__ (self, operator, op1, op2):
        assert op1.arity == op2.arity, 'binary operation on connection-sets with different arity'
        BinaryCSet.__init__ (self, operator, op1, op2)


class CSetIntersection (BinaryCSet):
    def __init__ (self, op1, op2):
        assert isinstance (op2, Mask), 'expected Mask operand'
        BinaryCSet.__init__ (self, "*", op1, op2)
        self._mask = op1.mask ().intersection (op2)

    def iterator (self, low0, high0, low1, high1, state):
        iter1 = self.op1.iterator (low0, high0, low1, high1, state)
        iter2 = self.op2.iterator (low0, high0, low1, high1, state)
        (i1, j1, v1) = next (iter1)
        (i2, j2) = next (iter2)
        while True:
            if (j1, i1) < (j2, i2):
                (i1, j1, v1) = next (iter1)
            elif (j2, i2) < (j1, i1):
                (i2, j2) = next (iter2)
            else:
                yield (i1, j1, v1)
                (i1, j1, v1) = next (iter1)
                (i2, j2) = next (iter2)


class CSetMultisetSum (BinaryCSets):
    def __init__ (self, op1, op2):
        BinaryCSet.__init__ (self, "+", op1, op2)
        self._mask = op1.mask ().multisetSum (op2.mask ())
        
    def iterator (self, low0, high0, low1, high1, state):
        iter1 = self.op1.iterator (low0, high0, low1, high1, state)
        iter2 = self.op2.iterator (low0, high0, low1, high1, state)
        try:
            (i1, j1, v1) = next (iter1)
        except StopIteration:
            (i2, j2, v2) = next (iter2)
            while True:
                yield (i2, j2, v2)
                (i2, j2, v2) = next (iter2)
        try:
            (i2, j2, v2) = next (iter2)
        except StopIteration:
            while True:
                yield (i1, j1, v1)
                (i1, j1, v1) = next (iter1)
        while True:
            i1s = i1
            j1s = j1
            while (j1, i1) <= (j2, i2):
                yield (i1, j1, v1)
                try:
                    (i1, j1, v1) = next (iter1)
                except StopIteration:
                    while True:
                        yield (i2, j2, v2)
                        (i2, j2, v2) = next (iter2)
            while (j2, i2) <= (j1s, i1s):
                yield (i2, j2, v2)
                try:
                    (i2, j2, v2) = next (iter2)
                except StopIteration:
                    while True:
                        yield (i1, j1, v1)
                        (i1, j1, v1) = next (iter1)

    def intersection (self, other):
        assert isinstance (other, Mask), 'expected Mask operand'
        if isFinite (self) or isFinite (other):
            # since operands are finite we are allowed to use isEmpty
            if isEmpty (self.op2.mask ().intersection (other)):
                return self.op1.intersection (other)
            if isEmpty (self.op1.mask ().intersection (other)):
                return self.op2.intersection (other)
        return CSetIntersection (self, other)


class TransposedMask (Finite, Mask):
    def __init__ (self, mask):
        self.subMask = mask

    def transpose (self):
        return self.subMask

    def bounds (self):
        (low0, high0, low1, high1) = self.subMask.bounds ()
        return (low1, high1, low0, high0)

    def startIteration (self, state):
        obj = copy.copy (self)
        obj.transposedState = state.transpose ()
        obj.subMask = self.subMask.startIteration (obj.transposedState)
        return obj

    def iterator (self, low0, high0, low1, high1, state):
        ls = []
        for c in self.subMask.iterator (low1, high1, low0, high0, \
                                        self.transposedState):
            ls.append ((c[1], c[0]))
        ls.sort (key=cmp_to_key(cmpPostOrder))
        return iter (ls)


class ShiftedMask (Mask):
    def __init__ (self, mask, M, N):
        self.subMask = mask
        self.M = M
        self.N = N

    def startIteration (self, state):
        obj = copy.copy (self)
        obj.subMask = self.subMask.startIteration (state)
        return obj

    def iterator (self, low0, high0, low1, high1, state):
        low0 -= self.M
        high0 -= self.M
        low1 -= self.N
        high1 -= self.N
        for (i, j) in self.subMask.iterator (max (low0, 0), high0, \
                                             max (low1, 0), high1, \
                                             state):
            (i1, j1) = (i + self.M, j + self.N)
            if i1 >= 0 and j1 >= 0:
                yield (i1, j1)


class FiniteShiftedMask (Finite, ShiftedMask):
    def bounds (self):
        (low0, high0, low1, high1) = self.subMask.bounds ()
        low0 += self.M
        high0 += self.M
        low1 += self.N
        high1 += self.N
        return (max (low0, 0), high0, max (low1, 0), high1)


def shiftedMask (mask, M, N):
    if isFinite (mask):
        return FiniteShiftedMask (mask, M, N)
    else:
        return ShiftedMask (mask, M, N)


class State (dict):
    def transpose (self):
        if 'partitions' in self:
            s = State (self)
            s['partitions'] = list (map (transpose, s['partitions']))
            return s
        else:
            return self


class MaskPartition (Finite, Mask):
    def __init__ (self, mask, partitions, selected, seed):
        Mask.__init__ (self)

        #*fixme* How can we know when this is not necessary?
        self.subMask = partitions[selected] * mask

        #domain = IntervalSetMask ([], [])
        #for m in partitions:
        #    assert isFinite (m), 'partitions must be finite'
        #    domain = domain.multisetSum (m)
        
        self.state = { #'domain' : domain,
                       'partitions' : partitions,
                       'selected' : selected }
        if seed != None:
            self.state['seed'] = seed

    def bounds (self):
        return self.subMask.bounds ()

    def startIteration (self, state):
        for key in self.state:
            state[key] = self.state[key]
        return self.subMask.startIteration (state)

    def iterator (self, low0, high0, low1, high1, state):
        raise RuntimeError ('iterator called on wrong object')


class CSetPartition (CSet):
    def __init__ (self, c, partitions, selected, seed):
        #*fixme* How can we know when this is not necessary?
        
        self.subCSet = (partitions[selected] * c).c
        CSet.__init__ (self, self.subCSet.mask (), *self.subCSet.valueSets)

        self.state = { #'domain' : domain,
                       'partitions' : partitions,
                       'selected' : selected }
        if seed != None:
            self.state['seed'] = seed

    def makeFiniteValueSet (self, k, bounds):
        return self.subCSet.makeFiniteValueSet (k, bounds);

    def bounds (self):
        return self.subCSet.bounds ()

    def startIteration (self, state):
        for key in self.state:
            state[key] = self.state[key]
        return self.subCSet.startIteration (state)

    def iterator (self, low0, high0, low1, high1, state):
        raise RuntimeError ('iterator called on wrong object')
    



#
#  This file is part of the Connection-Set Algebra (CSA).
#  Copyright (C) 2010,2011,2012,2019 Mikael Djurfeldt
#
#  CSA is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  CSA is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import math as _math
import random as _random
import numpy as _numpy

from . import intervalset as _iset

def grid2d (width, xScale = 1.0, yScale = 1.0, x0 = 0.0, y0 = 0.0):
    xScale /= width
    yScale /= width
    g = lambda i: \
        (x0 + xScale * (i % width), y0 + yScale * (i // width))
    g.type = 'grid'
    g.width = width
    g.xScale = xScale
    g.yScale = yScale
    g.x0 = x0
    g.y0 = y0
    g.inverse = lambda x, y: \
                    int (round (x / xScale - x0)) \
                    + width * int (round (y / yScale - y0))
    return g

def random2d (N, xScale = 1.0, yScale = 1.0):
    coords = [(xScale * _random.random (), yScale * _random.random ())
              for i in range (0, N)]
    g = lambda i: coords[i]
    g.type = 'ramdom'
    g.N = N
    g.xScale = xScale
    g.yScale = yScale
    # We should use a KD-tree here
    g.inverse = lambda x, y, domain=_iset.IntervalSet ((0, N - 1)): \
                    _numpy.array ([euclidDistance2d ((x, y), g(i)) \
                                   for i in domain]).argmin () \
                                   + domain.min ()
    return g

class ProjectionOperator (object):
    def __init__ (self, projection):
        self.projection = projection

    def __mul__ (self, g):
        projection = self.projection
        return lambda i: projection (g (i))

def euclidDistance2d (p1, p2):
    dx = p1[0] - p2[0]
    dy = p1[1] - p2[1]
    return _math.sqrt (dx * dx + dy * dy)

def euclidMetric2d (g1, g2 = None):
    g2 = g1 if g2 == None else g2
    return lambda i, j: euclidDistance2d (g1 (i), g2 (j))

# These functions were contributed by Dr. Birgit Kriener

def euclidToroidDistance2d (p1, p2, xScale=1.0, yScale=1.0):
    ddx, ddy = abs (p1[0] - p2[0]), abs (p1[1] - p2[1])
    dx = ddx if ddx < xScale/2. else  xScale - ddx
    dy = ddy if ddy < yScale/2. else  yScale - ddy
    return _math.sqrt (dx * dx + dy * dy)

def euclidToroidMetric2d (g1, g2 = None, xScale=1.0, yScale=1.0):
    g2 = g1 if g2 == None else g2
    return lambda i, j: euclidToroidDistance2d (g1 (i), g2 (j), xScale, yScale)

# 3D functions

def grid3d(width, xScale = 1.0, yScale = 1.0, zScale = 1.0, x0 = 0.0, y0 = 0.0, z0 = 0.0):
    """Returns a 3D grid between (0, 0, 0) and (1, 1, 1)
    :param width: The number of rows/columns the grid has
    :type width: int
    :param xScale: Scales the grid along the x axis
    :type xScale: float
    :param yScale: Scales the grid along the y axis
    :type yScale: float
    :param zScale: Scales the grid along the z axis
    :type zScale: float
    :param x0: Translates the grid along the x axis
    :type xScale: float
    :param y0: Translates the grid along the y axis
    :type yScale: float
    :param z0: Translates the grid along the z axis
    :type zScale: float
    :return: A callable grid that returns 3d positions when given an index"""
    xScale /= width
    yScale /= width
    zScale /= width
    g = lambda i: \
        (x0 + xScale * (i % width), y0 + yScale * ((i % (width*width)) / width), z0 + zScale * (i / (width*width)))
    g.type = 'grid3d'
    g.width = width
    g.xScale = xScale
    g.yScale = yScale
    g.zScale = zScale
    g.x0 = x0
    g.y0 = y0
    g.z0 = z0
    g.inverse = lambda x, y, z: \
                    int (round (x / xScale - x0)) \
                    + width * (int (round (y / yScale - y0)
                    + width * int (round (z / zScale - z0))))
    return g
    
def random3d(N, xScale = 1.0, yScale = 1.0, zScale = 1.0):
    """Creates a set of points scattered uniformly inside a 3D box
    :param N: Number of 3D points
    :type N: int
    :param xScale: The scale of the box on the x axis
    :type xScale: float
    :param yScale: The scale of the box on the y axis
    :type yScale: float
    :param zScale: The scale of the box on the z axis
    :type zScale: float
    """
    coords = _numpy.random.random((N, 3))
    coords[...,0] *= xScale
    coords[...,1] *= yScale
    coords[...,2] *= zScale
    g = lambda i: coords[i]
    g.type = 'random'
    g.N = N
    g.xScale = xScale
    g.yScale = yScale
    g.zScale = zScale
    g.inverse = lambda x, y, z, domain=_iset.IntervalSet ((0, N - 1)): \
                    _numpy.array ([euclidDistance3d (_numpy.array((x, y, z)), g(i)) \
                                   for i in domain]).argmin () \
                                   + domain.min ()
    return g

def euclidDistance3d(p1, p2):
    """Returns the euclidean distance in 3D between two points
    :param p1: The first point
    :type p1: numpy.array((3))
    :param p2: The second point
    :type p2: numpy.array((3))
    :return: The euclidean distance
    :rtype: float
    """
    return _numpy.linalg.norm(p2 - p1)

def euclidMetric3d (g1, g2 = None):
    """Returns an euclidean metric for 3D points
    :param g1: The first group of points
    :type g1: callable
    :param g2: The second group of points. If None, the first group of points is used
    :type g2: callable
    :return: A 3D euclidean metric function
    :rtype: function
    """
    g2 = g1 if g2 == None else g2
    return lambda i, j: euclidDistance3d (g1 (i), g2 (j))



#
#  This file is part of the Connection-Set Algebra (CSA).
#  Copyright (C) 2010,2011,2012 Mikael Djurfeldt
#
#  CSA is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  CSA is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import sys

from .csaobject import *

infinity = sys.maxsize - 1

# Interval sets are represented as ordered lists of closed intervals
#
class IntervalSet (CSAObject):
    tag = 'intervalset'
    
    @staticmethod
    # return true if tuple i represents a well-formed interval
    def goodInterval (i):
        return len (i) == 2 \
               and isinstance (i[0], int) \
               and isinstance (i[1], int) \
               and i[0] <= i[1]

    @staticmethod
    def rangeToIntervals (x):
        if not x:
            return []
        elif len (x) == 1:
            return [(x[0], x[0])]
        elif x[1] - x[0] == 1:
            return [(x[0], x[-1])]
        else:
            return ((e, e) for e in x)

    @staticmethod
    def coerce (s):
        if not isinstance (s, list):
            s = [ s ]
            
        res = []
        for x in s:
            if isinstance (x, tuple):
                assert IntervalSet.goodInterval (x), 'malformed interval'
                res.append (x)
            elif isinstance (x, int):
                res.append ((x, x))
            elif isinstance (x, range):
                res += IntervalSet.rangeToIntervals (x)
            else:
                raise TypeError ("can't interpret element as interval")
        s = res

        s.sort ()

        # merge intervals
        # by construction we know that i[0] <= i[1] for i in s
        res = []
        N = 0
        if s:
            lastLower = s[0][0]
            lastUpper = s[0][1]

            assert lastLower >= 0, 'only positive values allowed'

            for i in s[1:]:
                assert lastLower < i[0] and lastUpper < i[0], 'intervals overlap'
                if i[0] - lastUpper == 1:
                    lastUpper = i[1]
                else:
                    res.append ((lastLower, lastUpper))
                    N += 1 + lastUpper - lastLower
                    lastLower = i[0]
                    lastUpper = i[1]
            res.append ((lastLower, lastUpper))
            N += 1 + lastUpper - lastLower

        return (res, N)
    
    def __init__ (self, s = [], intervals = None, nIntegers = None):
        if intervals:
            self.intervals = intervals
            self.nIntegers = nIntegers
        else:
            (self.intervals, self.nIntegers) = self.coerce (s)

    def repr (self):
        return 'IntervalSet(%r)' % self.intervals

    def __len__ (self):
        return self.nIntegers

    def __contains__ (self, n):
        for i in self.intervals:
            if n > i[1]:
                continue
            elif n >= i[0]:
                return True
            else:
                return False
        return False

    def __iter__ (self):
        for i in self.intervals:
            for e in range (i[0], i[1] + 1):
                yield e

    def __invert__ (self):
        return ComplementaryIntervalSet (intervals = self.intervals, \
                                         nIntegers = self.nIntegers)

    def __add__ (self, other):
        if not isinstance (other, IntervalSet):
            other = IntervalSet (other)
        return self.union (other)

    def __radd__ (self, other):
        return IntervalSet (other).union (self)

    def __sub__ (self, other):
        if not isinstance (other, IntervalSet):
            other = IntervalSet (other)
        return self.intersection (~other)

    def __rsub__ (self, other):
        return IntervalSet (other).intersection (~self)

    def __mul__ (self, other):
        if not isinstance (other, IntervalSet):
            other = IntervalSet (other)
        return self.intersection (other)

    def __rmul__ (self, other):
        return IntervalSet (other).intersection (self)

    def finite (self):
        return True

    def shift (self, N):
        if not self or N == 0:
            return self

        intervals = []
        nIntegers = self.nIntegers
        for (i, j) in self.intervals:
            i += N
            j += N
            if i >= 0:
                intervals.append ((i, j))
            elif j >= 0:
                intervals.append ((0, j))
                nIntegers += i
                
        return IntervalSet (intervals = intervals, nIntegers = nIntegers)

    def intervalIterator (self):
        return iter (self.intervals)

    def boundedIterator (self, low, high):
        iterator = iter (self.intervals)
        i = next (iterator)
        while i[1] < low:
            i = next (iterator)
        while i[0] < high:
            for e in range (max (low, i[0]), min (i[1] + 1, high)):
                yield e
            i = next (iterator)

    def count (self, low, high):
        iterator = iter (self.intervals)
        c = 0
        try:
            i = next (iterator)
            while i[1] < low:
                i = next (iterator)
            while i[0] < high:
                c += min (i[1] + 1, high) - max (low, i[0])
                i = next (iterator)
        except StopIteration:
            pass
        return c

    def min (self):
        return self.intervals[0][0]

    def max (self):
        return self.intervals[-1][1]

    def skipIntervals (self):
        if len (self.intervals) <= 1 or self.intervals[0][0] != self.intervals[0][1]:
            return 1, self.intervals
        skip = self.intervals[1][0] - self.intervals[0][0]
        res = []
        start = last = self.intervals[0][0]
        for i in self.intervals[1:]:
            if i[0] != i[1]:
                return 1, self.intervals
            if i[0] != last + skip:
                if i[0] % skip != 0:
                    return 1, self.intervals
                res.append ((start, last))
                start = i[0]
            last = i[0]
        res.append ((start, last))
        return skip, res

    def intersection (self, other):
        res = []
        N = 0
        iter0 = self.intervalIterator ()
        iter1 = other.intervalIterator ()
        try:
            i0 = next (iter0)
            i1 = next (iter1)
            while True:
                if i0[1] <= i1[1]:
                    if i0[1] >= i1[0]:
                        lower = max (i0[0], i1[0])
                        res.append ((lower, i0[1]))
                        N += 1 + i0[1] - lower
                    i0 = next (iter0)
                else:
                    if i1[1] >= i0[0]:
                        lower = max (i0[0], i1[0])
                        res.append ((lower, i1[1]))
                        N += 1 + i1[1] - lower
                    i1 = next (iter1)
        except StopIteration:
            pass
        iset = IntervalSet ()
        iset.intervals = res
        iset.nIntegers = N
        return iset

    def union (self, other):
        if isinstance (other, ComplementaryIntervalSet):
            return ~(~self).intersection (~other)
        
        iset = IntervalSet ()
        if not other.nIntegers:
            iset.intervals = list (self.intervals)
            iset.nIntegers = self.nIntegers
            return iset
        if not self.nIntegers:
            iset.intervals = list (other.intervals)
            iset.nIntegers = other.nIntegers
            return iset
        res = []
        N = 0
        iter0 = self.intervalIterator ()
        iter1 = other.intervalIterator ()
        i0 = next (iter0)
        i1 = next (iter1)
        if i0[0] <= i1[0]:
            (lower, upper) = i0
        else:
            (lower, upper) = i1
        try:
            while True:
                if i0[0] <= i1[0]:
                    if i0[0] <= upper + 1:
                        if i0[1] > upper:
                            upper = i0[1]
                    else:
                        res.append ((lower, upper))
                        N += 1 + upper - lower
                        (lower, upper) = i0
                    try:
                        i0 = next (iter0)
                    except StopIteration:
                        if i1[0] <= upper + 1:
                            if i1[1] > upper:
                                upper = i1[1]
                            i1 = (lower, upper)
                        else:
                            res.append ((lower, upper))
                            N += 1 + upper - lower
                        while True:
                            res.append (i1)
                            N += 1 + i1[1] - i1[0]
                            i1 = next (iter1)
                else:
                    if i1[0] <= upper + 1:
                        if i1[1] > upper:
                            upper = i1[1]
                    else:
                        res.append ((lower, upper))
                        N += 1 + upper - lower
                        (lower, upper) = i1
                    try:
                        i1 = next (iter1)
                    except StopIteration:
                        if i0[0] <= upper + 1:
                            if i0[1] > upper:
                                upper = i0[1]
                            i0 = (lower, upper)
                        else:
                            res.append ((lower, upper))
                            N += 1 + upper - lower
                        while True:
                            res.append (i0)
                            N += 1 + i0[1] - i0[0]
                            i0 = next (iter0)
        except StopIteration:
            pass
        iset.intervals = res
        iset.nIntegers = N
        return iset

    def _to_xml (self):
        intervals = [ E ('interval', E ('cn', str (i)), E ('cn', str (j)))
                      for (i, j) in self.intervals ]
        return E (IntervalSet.tag, *intervals)

    @classmethod
    def from_xml (cls, element, env = {}):
        intervals = []
        for ivalElement in element.getchildren ():
            ival = ivalElement.getchildren ()
            intervals.append ((int (ival[0].text), int (ival[1].text)))
        return IntervalSet (intervals)

CSAObject.tag_map[CSA + IntervalSet.tag] = (IntervalSet, CUSTOM)


class ComplementaryIntervalSet (IntervalSet):
    def __init__ (self, s = [], intervals = None, nIntegers = None):
        IntervalSet.__init__ (self, s, intervals, nIntegers)

    def repr (self):
        if not self.intervals:
            return 'N'
        else:
            return '~IntervalSet(%r)' % self.intervals

    def __bool__ (self):
        return True

    # def __len__ (self):
    #     raise RuntimeError ('ComplementaryIntervalSet has infinite length')
 
    def __contains__ (self, n):
        for i in self.intervals:
            if n < i[0]:
                continue
            elif n <= i[1]:
                return False
            else:
                return True
        return True

    def __iter__ (self):
        raise RuntimeError ("can't interate over ComplementaryIntervalSet")

    def __invert__ (self):
        return IntervalSet (intervals = self.intervals, \
                            nIntegers = self.nIntegers)

    def finite (self):
        return False

    def shift (self, N):
        iset = IntervalSet (intervals = self.intervals, \
                            nIntegers = self.nIntegers).shift (N)
        return ComplementaryIntervalSet (intervals = iset.intervals, \
                                         nIntegers = iset.nIntegers)

    def intervalIterator (self):
        start = 0
        for i in self.intervals:
            if i[0] > 0:
                yield (start, i[0] - 1)
            start = i[1] + 1
        yield (start, infinity)

    def boundedIterator (self, low, high):
        raise RuntimeError ("can't interate over ComplementaryIntervalSet")

    def count (self, low, high):
        iterator = iter (self.intervals)
        c = 0
        prev = low
        try:
            i = next (iterator)
            while i[1] < low:
                i = next (iterator)
            while i[0] < high:
                c += i[0] - prev
                prev = i[1] + 1
                i = next (iterator)
        except StopIteration:
            pass
        if prev < high:
            c += high - prev
        return c

    def min (self):
        if not self.intervals or self.intervals[0][0] > 0:
            return 0
        else:
            return self.intervals[0][1] + 1

    def max (self):
        raise RuntimeError ('the maximum of a ComplementaryIntervalSet is infinity')

    def intersection (self, other):
        if isinstance (other, ComplementaryIntervalSet):
            return ~(~self).union (~other)
        else:
            return IntervalSet.intersection (self, other)

    def union (self, other):
        return ~(~self).intersection (~other)

    def _to_xml (self):
        if not self.intervals:
            return E ('N')
        else:
            return E ('apply', E ('complement'), IntervalSet._to_xml (self))


N = ComplementaryIntervalSet ([])

CSAObject.tag_map[CSA + 'N'] = (N, SINGLETON)



#
#  This file is part of the Connection-Set Algebra (CSA).
#  Copyright (C) 2010,2011,2012,2019 Mikael Djurfeldt
#
#  CSA is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  CSA is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import math
import random
import copy
#from scipy.spatial import KDTree

from . import connset as cs
from . import valueset as vs
from . import _elementary

from .csaobject import *

class Random (cs.Operator):
    def __mul__ (self, valueSet):
        return ValueSetRandomMask (valueSet)
    
    def __call__ (self, p = None, N = None, fanIn = None, fanOut = None):
        if p != None:
            assert N == None and fanIn == None and fanOut == None, \
                   'inconsistent parameters'
            return _elementary.ConstantRandomMask (p)
        elif N != None:
            assert fanIn == None and fanOut == None, \
                   'inconsistent parameters'
            return _elementary.SampleNRandomOperator (N)
        elif fanIn != None:
            assert fanOut == None, \
                   'inconsistent parameters'
            return _elementary.FanInRandomOperator (fanIn)
        elif fanOut != None:
            return _elementary.FanOutRandomOperator (fanOut)
        assert False, 'inconsistent parameters'


class ValueSetRandomMask (cs.Mask):
    def __init__ (self, valueSet):
        cs.Mask.__init__ (self)
        self.valueSet = valueSet
        self.state = random.getstate ()

    def startIteration (self, state):
        random.setstate (self.state)
        return self

    def iterator (self, low0, high0, low1, high1, state):
        for j in range (low1, high1):
            for i in range (low0, high0):
                if random.random () < self.valueSet (i, j):
                    yield (i, j)

    def _to_xml (self):
        return CSAObject.apply ('times', 'random', self.valueSet._to_xml ())


class Disc (cs.Operator):
    def __init__ (self, r):
        self.r = r

    def __mul__ (self, metric):
        return DiscMask (self.r, metric)


class DiscMask (cs.Mask):
    def __init__ (self, r, metric):
        cs.Mask.__init__ (self)
        self.r = r
        self.metric = metric

    def iterator (self, low0, high0, low1, high1, state):
        for j in range (low1, high1):
            for i in range (low0, high0):
                if self.metric (i, j) < self.r:
                    yield (i, j)


class Rectangle (cs.Operator):
    def __init__ (self, width, height):
        self.width = width
        self.height = height

    def __mul__ (self, gFunction):
        if isinstance (gFunction, tuple):
            return RectangleMask (self.width, self.height,
                                  gFunction[0], gFunction[1])
        else:
            return RectangleMask (self.width, self.height, gFunction, gFunction)


class RectangleMask (cs.Mask):
    def __init__ (self, width, height, g0, g1):
        cs.Mask.__init__ (self)
        self.hwidth = width / 2.0
        self.hheight = height / 2.0
        self.g0 = g0
        self.g1 = g1

    def iterator (self, low0, high0, low1, high1, state):
        for j in range (low1, high1):
            for i in range (low0, high0):
                p0 = self.g0 (i)
                p1 = self.g1 (j)
                dx = p0[0] - p1[0]
                dy = p0[1] - p1[1]
                if abs (dx) < self.hwidth and abs (dy) < self.hheight:
                    yield (i, j)

 
class Gaussian (cs.Operator):
    def __init__ (self, sigma, cutoff):
        cs.Operator.__init__ (self, 'gaussian')
        self.sigma = sigma
        self.cutoff = cutoff
        
    def __mul__ (self, metric):
        return GaussianValueSet (self, metric)


class GaussianValueSet (OpExprValue, vs.ValueSet):
    def __init__ (self, operator, metric):
        OpExprValue.__init__ (self, operator, metric)
        self.sigma22 = 2 * operator.sigma * operator.sigma
        self.cutoff = operator.cutoff
        self.metric = metric

    def __call__ (self, i, j):
        d = self.metric (i, j)
        return math.exp (- d * d / self.sigma22) if d < self.cutoff else 0.0


class Block (cs.Operator):
    def __init__ (self, M, N):
        self.M = M
        self.N = N

    def __mul__ (self, other):
        c = cs.coerceCSet (other)
        if isinstance (c, cs.Mask):
            return BlockMask (self.M, self.N, c)
        else:
            return cs.ConnectionSet (BlockCSet (self.M, self.N, c))


class BlockMask (cs.Mask):
    def __init__ (self, M, N, mask):
        cs.Mask.__init__ (self)
        self.M = M
        self.N = N
        self.m = mask

    def startIteration (self, state):
        #*fixme* filter out 'partitions' from state
        nState = {}
        for k in state:
            if k != 'partitions':
                nState[k] = state[k]
        self.obj = self.m.startIteration (nState)
        return self

    def iterator (self, low0, high0, low1, high1, state):
        maskIter =  self.obj.iterator (low0 // self.M,
                                       (high0 + self.M - 1) // self.M,
                                       low1 // self.N,
                                       (high1 + self.N - 1) // self.N,
                                       state)
        try:
            pre = []
            (i, j) = next (maskIter)
            while True:
                # collect connections in one connection matrix column
                post = j
                while j == post:
                    pre.append (i)
                    (i, j) = next (maskIter)

                # generate blocks for the column
                for jj in range (max (self.N * post, low1),
                                  min (self.N * (post + 1), high1)):
                    for k in pre:
                        for ii in range (max (self.M * k, low0),
                                          min (self.M * (k + 1), high0)):
                            yield (ii, jj)
                pre = []
        except StopIteration:
            if pre:
                # generate blocks for the last column
                for jj in range (max (self.N * post, low1),
                                  min (self.N * (post + 1), high1)):
                    for k in pre:
                        for ii in range (max (self.M * k, low0),
                                          min (self.M * (k + 1), high0)):
                            yield (ii, jj)


class Repeat (cs.Operator):
    def __init__ (self, M, N):
        self.M = M
        self.N = N

    def __mul__ (self, other):
        c = cs.coerceCSet (other)
        if isinstance (c, cs.Mask):
            return RepeatMask (self.M, self.N, c)
        else:
            return cs.ConnectionSet (RepeatCSet (self.M, self.N, c))


# Not fully implemented
# Currently only handles cases where we iterate over an even number of
# ocurrences of the template mask, both with regard to sources and targets
#
class RepeatMask (cs.Mask):
    def __init__ (self, M, N, mask):
        cs.Mask.__init__ (self)
        self.M = M
        self.N = N
        self.m = mask

    def iterator (self, low0, high0, low1, high1, state):
        jj = low1
        nextHigh1 = (low1 + self.N) / self.N * self.N
        while nextHigh1 <= high1:
            maskIter =  self.m.iterator (0,
                                         self.M,
                                         0,
                                         self.N,
                                         state)
            try:
                (i, j) = next (maskIter)
                post = j
                while post < self.N:
                    pre = []
                    while j == post:
                        pre.append (i)
                        (i, j) = next (maskIter)
                    ii = low0
                    while ii < high0:
                        for k in pre:
                            yield (ii + k, jj + post)
                        ii += self.M
                    post = j
            except StopIteration:
                ii = low0
                while ii < high0:
                    for k in pre:
                        yield (ii + k, jj + post)
                    ii += self.M
            jj = nextHigh1
            nextHigh1 += self.N


class Transpose (cs.Operator):
    def __mul__ (self, other):
        c = cs.coerceCSet (other)
        if isinstance (c, cs.Mask):
            return other.transpose ()
        else:
            return cs.ConnectionSet (other.transpose ())


class Shift (cs.Operator):
    def __init__ (self, M, N):
        self.M = M
        self.N = N

    def __mul__ (self, other):
        c = cs.coerceCSet (other)
        if isinstance (c, cs.Mask):
            return other.shift (self.M, self.N)
        else:
            return cs.ConnectionSet (other.shift (self.M, self.N))


class Fix (cs.Operator):
    def __mul__ (self, other):
        c = cs.coerceCSet (other)
        if isinstance (c, cs.Mask):
            return FixedMask (other)
        else:
            return cs.ConnectionSet (FixedCSet (other))


class FixedMask (cs.FiniteMask):
    def __init__ (self, mask):
        cs.FiniteMask.__init__ (self)
        ls = []
        for c in mask:
            ls.append (c)
        self.connections = ls
        targets = list (map (cs.target, ls))
        self.low0 = min (ls)[0]
        self.high0 = max (ls)[0] + 1
        self.low1 = min (targets)
        self.high1 = max (targets) + 1

    def iterator (self, low0, high0, low1, high1, state):
        if not self.isBoundedBy (low0, high0, low1, high1):
            return iter (self.connections)
        else:
            return self.boundedIterator (low0, high0, low1, high1)

    def boundedIterator (self, low0, high0, low1, high1):
        for c in self.connections:
            if low0 <= c[0] and c[0] < high0 \
               and low1 <= c[1] and c[1] < high1:
                yield c



#
#  This file is part of the Connection-Set Algebra (CSA).
#  Copyright (C) 2010,2011,2012 Mikael Djurfeldt
#
#  CSA is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  CSA is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

try:
    from nineml.connection_generator import ConnectionGenerator
    HAVE_CG=True
except ImportError:
    HAVE_CG=False

if HAVE_CG:
    from .csaobject import from_xml
    from .elementary import arity, cross, partition
    from .closure import Closure
    
    class CSAConnectionGenerator (ConnectionGenerator):
        def __init__ (self, cset):
            self.cset = cset
            self.generator = False

        @property
        def arity (self):
            return arity (self.cset)

        def setMask (self, mask):
            self.setMasks ([mask], 0)

        def setMasks (self, masks, local):
            csaMasks = list (map (CSAConnectionGenerator.makeMask, masks))
            self.generator = partition (self.cset, csaMasks, local)

        @staticmethod
        def makeMask (mask):
            return cross (CSAConnectionGenerator.makeIList (mask.sources),
                          CSAConnectionGenerator.makeIList (mask.targets))

        @staticmethod
        def makeIList (iset):
            if iset.skip == 1:
                return iset.intervals
            else:
                ls = []
                for ivl in iset.intervals:
                    for i in range (ivl[0], ivl[1] + 1, iset.skip):
                        ls.append ((i, i))
                return ls

        def __len__ (self):
            return self.generator.__len__ ()

        def __iter__ (self):
            return self.generator.__iter__ ()

def connectionGeneratorClosureFromXML (element):
    cset = from_xml (element)
    if isinstance (cset, Closure):
        return lambda *args: CSAConnectionGenerator (cset (*args))
    else:
        return lambda: CSAConnectionGenerator (cset)



#
#  This file is part of the Connection-Set Algebra (CSA).
#  Copyright (C) 2010,2011,2012,2017 Mikael Djurfeldt
#
#  CSA is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  CSA is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import ctypes
try:
    _libpycsa_handle_ = ctypes.CDLL ('/usr/local/lib/libpycsa.so')
except OSError:
    None

from .version import __version__
from .connset import Mask
from .connset import ConnectionSet
from .elementary import *
#from operators import *
#from arithmetic import *
from .misc import *
from .geometry import *
from .csaobject import parse, from_xml
from .plot import *
from .closure import *
from .conngen import *



#
#  This file is part of the Connection-Set Algebra (CSA).
#  Copyright (C) 2010,2011,2012 Mikael Djurfeldt
#
#  CSA is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  CSA is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from . import _misc

def disc (r):
    return _misc.Disc (r)

def rectangle (width, height):
    return _misc.Rectangle (width, height)
 
def gaussian (sigma, cutoff):
    return _misc.Gaussian (sigma, cutoff)

def block (M, N = None):
    return _misc.Block (M, M if N == None else N)

def repeat (M, N = None):
    return _misc.Repeat (M, M if N == None else N)

transpose = _misc.Transpose ()

def shift (M, N):
    return _misc.Shift (M, N)

fix = _misc.Fix ()

def block1 (N):
    return _misc.Block (N)

#del _misc                               # not for export



#
#  This file is part of the Connection-Set Algebra (CSA).
#  Copyright (C) 2010,2011,2012 Mikael Djurfeldt
#
#  CSA is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  CSA is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import print_function
import sys as _sys

from . import intervalset as _iset
from . import connset as _cs
from . import valueset as _vs
from . import _elementary
from . import _misc
from .csaobject import registerTag

# Connection-Set constructor
#
def cset (mask, *valueSets):
    if valueSets:
        c = _cs.ExplicitCSet (mask, *valueSets)
        return _cs.ConnectionSet (c)
    else:
        return mask

registerTag (_cs.CSet.tag, cset, 1)

# Selectors
#
def mask (obj):
    cset = _cs.coerceCSet (obj)
    return cset.mask ()

def value (obj, k):
    assert isinstance (obj, _cs.ConnectionSet), 'expected connection-set'
    return obj.c.value (k)

def arity (obj):
    if isinstance (obj, _cs.ConnectionSet):
        return obj.c.arity
    else:
        return 0

# Value-set constructor
#
def vset (obj):
    if not callable (obj):
        return _vs.QuotedValueSet (obj)
    else:
        return _vs.GenericValueSet (obj)

# Intervals
#
def ival (beg, end):
    return _iset.IntervalSet ((beg, end))

N = _iset.N

# Cartesian product
#
def cross (set0, set1):
    return _cs.intervalSetMask (set0, set1)

# Elementary masks
#
empty = cross ([], [])

full = _elementary.FullMask ()

oneToOne = _elementary.OneToOne ()

random = _misc.Random ()

# Support for parallel simulator
#
def partition (c, masks, selected, seed = None):
    if isinstance (c, _cs.Mask):
        return _cs.MaskPartition (c, masks, selected, seed)
    elif isinstance (c, _cs.ConnectionSet):
        return _cs.ConnectionSet (_cs.CSetPartition (c, masks, selected, seed))

# Utilities
#
def tabulate (c):
    for x in c:
        print(u'{}'.format(x[0]), end=u' ')
        for e in x[1:]:
            print(u'\t{}'.format(e))
        
#del _elementary, cs, sys                # not for export
